/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "lua.pmh"

namespace fe
{

LuaRecordType::LuaRecordType(LuaContext *pContext,const String &mtn)
	: t_lua_record_t(pContext, mtn)
{
	setMethod("isValid", LuaRecordObject::isValid);
	setMethod("check", LuaRecordObject::check);
	setMethod("layout", LuaRecordObject::layout);
	setMethod("clone", LuaRecordObject::clone);
	//setMethod("deepclone", LuaRecordObject::deepclone);
}

void LuaRecordType::pushNewMetaTable(void)
{
	lua_State *pLuaState = m_pContext->state();

//	lua_newtable(pLuaState);
	luaL_newmetatable(pLuaState,m_mtname.c_str());

	lua_pushstring(pLuaState, "__gc");
	lua_pushcfunction(pLuaState, LuaRecordObject::destroy);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__index");
	lua_pushcfunction(pLuaState, LuaRecordObject::index);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__newindex");
	lua_pushcfunction(pLuaState, LuaRecordObject::newindex);
	lua_settable(pLuaState, -3);
}

int LuaRecordObject::index(lua_State *pLuaState)
{
	String attributeName = lua_tostring(pLuaState, -1);

	lua_pop(pLuaState, 1);
	LuaRecordObject *pObj = (LuaRecordObject *)lua_touserdata(pLuaState, -1);

	attributeName=pObj->aliasedName(attributeName);

	Instance instance;
	if(pObj->data().isValid() &&
			pObj->data().extractInstance(instance, attributeName))
	{
		pObj->context()->pushInstance(instance);
	}
	else
	{
		if(!pObj->type()->pushMethod(attributeName))
		{
			String s;
			s.sPrintf("unknown record attribute \'%s\'", attributeName.c_str());
			LuaContext::error(pLuaState, s.c_str());
			return 0;
		}
	}
	return 1;
}

int LuaRecordObject::newindex(lua_State *pLuaState)
{
	LuaRecordObject *pObj = (LuaRecordObject *)lua_touserdata(pLuaState, 1);

	String attributeName = lua_tostring(pLuaState, -2);
	attributeName=pObj->aliasedName(attributeName);

	Instance instance;
	if(pObj->data().extractInstance(instance, attributeName))
	{
// what in the world is this
#if 0
		LuaVector4Object* pOther;
		if((pOther=reinterpret_cast<LuaVector4Object*>
				(luaL_checkudata(pLuaState,2,"mt_vector4"))))
		{
		}
#endif

		// value is still at top of stack
		pObj->context()->popInstance(instance);

		// pop other tw
		lua_pop(pLuaState, 2);

		// push the return value
		pObj->context()->pushInstance(instance);
	}
	else
	{
		lua_pop(pLuaState, 3);
		String s;
		s.sPrintf("unknown record attribute \'%s\'", attributeName.c_str());
		LuaContext::error(pLuaState, s.c_str());
		return 0;
	}
	return 1;
}

int LuaRecordObject::isValid(lua_State *pLuaState)
{
	LuaRecordObject *pObj = (LuaRecordObject *)selfIsFirst(pLuaState);

	lua_pushboolean(pLuaState, pObj && pObj->data().isValid());
	return 1;
}

int LuaRecordObject::check(lua_State *pLuaState)
{
	LuaRecordObject *pObj = (LuaRecordObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	String attrName = lua_tostring(pLuaState, -1);

	attrName=pObj->aliasedName(attrName);

	lua_pushboolean(pLuaState,
			(int)pObj->data().layout()->checkAttributeStr(attrName));

	return 1;
}

int LuaRecordObject::layout(lua_State *pLuaState)
{
	LuaRecordObject *pObj = (LuaRecordObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	t_lua_layout::push(pObj->context(), pObj->data().layout());

	return 1;
}

int LuaRecordObject::clone(lua_State *pLuaState)
{
	LuaRecordObject *pObj = (LuaRecordObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	t_lua_record::push(pObj->context(), pObj->data().clone());

	return 1;
}

#if 0
int LuaRecordObject::deepclone(lua_State *pLuaState)
{
	LuaRecordObject *pObj = (LuaRecordObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	t_lua_record::push(pObj->context(), pObj->data().deepclone());

	return 1;
}
#endif

LuaLayoutType::LuaLayoutType(LuaContext *pContext,const String &mtn)
	: t_lua_layout_t(pContext, mtn)
{
	setMethod("scope", LuaLayoutObject::scope);
	setMethod("populate", LuaLayoutObject::populate);
	setMethod("createRecord", LuaLayoutObject::createRecord);
	setMethod("name", LuaLayoutObject::name);
}

int LuaLayoutObject::index(lua_State *pLuaState)
{
	String attributeName = lua_tostring(pLuaState, -1);

	lua_pop(pLuaState, 1);
	LuaLayoutObject *pObj = (LuaLayoutObject *)selfIsFirst(pLuaState);

	attributeName=pObj->aliasedName(attributeName);

	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	if(!pObj->type()->pushMethod(attributeName))
	{
		lua_pushnil(pLuaState);
	}
	return 1;
}

int LuaLayoutObject::scope(lua_State *pLuaState)
{
	LuaLayoutObject *pObj = (LuaLayoutObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	t_lua_scope::push(pObj->context(), pObj->data()->scope());

	return 1;
}

int LuaLayoutObject::populate(lua_State *pLuaState)
{
	LuaLayoutObject *pObj = (LuaLayoutObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	if(lua_gettop(pLuaState) == 3)
	{
		String attrName = lua_tostring(pLuaState, -2);
		String typeName = lua_tostring(pLuaState, -1);

		attrName=pObj->aliasedName(attrName);

		pObj->data()->populate(attrName,typeName);
	}
	else if(lua_gettop(pLuaState) == 2)
	{
		LuaBaseObject *pB = (LuaBaseObject *)lua_touserdata(pLuaState, -1);
		if(pB && (pB->type() == pObj->type()->context()->lookup(LuaContext::e_layout)))
		{
			LuaLayoutObject *pL = (LuaLayoutObject *)pB;
			pObj->data()->populate(pL->data());
		}
		else
		{
			sp<PopulateI> spPopulateI = component<PopulateI>(pLuaState, -1);
			spPopulateI->populate(pObj->data());
		}
	}

	return 0;
}

int LuaLayoutObject::createRecord(lua_State *pLuaState)
{
	LuaLayoutObject *pObj = (LuaLayoutObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	t_lua_record::push(pObj->context(),
			pObj->data()->scope()->createRecord(pObj->data()));

	return 1;
}

int LuaLayoutObject::name(lua_State *pLuaState)
{
	LuaLayoutObject *pObj = (LuaLayoutObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	lua_pushstring(pLuaState, pObj->data()->name().c_str());

	return 1;
}

LuaScopeType::LuaScopeType(LuaContext *pContext,const String &mtn)
	: t_lua_scope_t(pContext, mtn)
{
	setMethod("declare", LuaScopeObject::declare);
	setMethod("createRecord", LuaScopeObject::createRecord);
}

int LuaScopeObject::index(lua_State *pLuaState)
{
	String attributeName = lua_tostring(pLuaState, -1);

	lua_pop(pLuaState, 1);
	LuaScopeObject *pObj = (LuaScopeObject *)selfIsFirst(pLuaState);

	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	attributeName=pObj->aliasedName(attributeName);

	if(!pObj->type()->pushMethod(attributeName))
	{
		lua_pushnil(pLuaState);
	}
	return 1;
}

int LuaScopeObject::declare(lua_State *pLuaState)
{
	LuaScopeObject *pObj = (LuaScopeObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	String layoutName = lua_tostring(pLuaState, -1);

	t_lua_layout::push(pObj->context(), pObj->data()->declare(layoutName));

	return 1;
}

int LuaScopeObject::createRecord(lua_State *pLuaState)
{
	LuaScopeObject *pObj = (LuaScopeObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	LuaBaseObject *pB = (LuaBaseObject *)lua_touserdata(pLuaState, -1);
	if(!pB)
	{
		LuaContext::error(pLuaState, "invalid layout");
		return 0;
	}
	if(pB->type() != pObj->type()->context()->lookup(LuaContext::e_layout))
	{
		lua_pushnil(pLuaState);
		return 1;
	}

	LuaLayoutObject *pL = (LuaLayoutObject *)pB;

	t_lua_record::push(pObj->context(),pObj->data()->createRecord(pL->data()));

	return 1;
}

LuaGroupType::LuaGroupType(LuaContext *pContext,const String &mtn)
	: t_lua_group_t(pContext, mtn)
{
	setMethod("length", LuaGroupObject::length);
	setMethod("add", LuaGroupObject::add);
	setMethod("remove", LuaGroupObject::remove);
	setMethod("lookup", LuaGroupObject::lookup);
	setMethod("setWeak", LuaGroupObject::setWeak);
	setMethod("clear", LuaGroupObject::clear);
}

int LuaGroupObject::length(lua_State *pLuaState)
{
	LuaGroupObject *pObj = (LuaGroupObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	lua_pushnumber(pLuaState, pObj->data()->count());

	return 1;
}

int LuaGroupObject::add(lua_State *pLuaState)
{
	LuaGroupObject *pObj = (LuaGroupObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	LuaBaseObject *pB = (LuaBaseObject *)lua_touserdata(pLuaState, -1);
	if(!pB)
	{
		LuaContext::error(pLuaState, "invalid record");
		return 0;
	}
	if(pB->type() != pObj->type()->context()->lookup(LuaContext::e_record))
	{
		lua_pushnil(pLuaState);
		return 1;
	}

	LuaRecordObject *pR = (LuaRecordObject *)pB;

	pObj->data()->add(pR->data());

	lua_pushnil(pLuaState);
	return 1;
}

int LuaGroupObject::remove(lua_State *pLuaState)
{
	LuaGroupObject *pObj = (LuaGroupObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	LuaBaseObject *pB = (LuaBaseObject *)lua_touserdata(pLuaState, -1);
	if(!pB)
	{
		LuaContext::error(pLuaState, "invalid record");
		return 0;
	}
	if(pB->type() != pObj->type()->context()->lookup(LuaContext::e_record))
	{
		lua_pushnil(pLuaState);
		return 1;
	}

	LuaRecordObject *pR = (LuaRecordObject *)pB;

	pObj->data()->remove(pR->data());

	lua_pushnil(pLuaState);
	return 1;
}

int LuaGroupObject::lookup(lua_State *pLuaState)
{
	LuaGroupObject *pObj = (LuaGroupObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	LuaScopeObject *pS = to<LuaScopeObject>(pLuaState, 2, LuaContext::e_scope);
	sp<Scope> spScope = pS->data();

	String attr_name = lua_tostring(pLuaState, 3);
	String attr_value = lua_tostring(pLuaState, 4);

	Accessor<String> aName(spScope, attr_name);

	Record r_found = pObj->data()->lookup<String>(aName, attr_value);

	t_lua_record::push(pObj->context(), r_found);
	return 1;
}

int LuaGroupObject::setWeak(lua_State *pLuaState)
{
	LuaGroupObject *pObj = (LuaGroupObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	int weak = (int)lua_tonumber(pLuaState, 2);

	pObj->data()->setWeak(weak!=0);

	lua_pushnil(pLuaState);
	return 1;
}

int LuaGroupObject::clear(lua_State *pLuaState)
{
	LuaGroupObject *pObj = (LuaGroupObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }


	pObj->data()->clear();

	lua_pushnil(pLuaState);
	return 1;
}

int LuaGroupObject::index(lua_State *pLuaState)
{
	String attributeName = lua_tostring(pLuaState, -1);

	lua_pop(pLuaState, 1);
	LuaGroupObject *pObj = (LuaGroupObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	attributeName=pObj->aliasedName(attributeName);

	if(!pObj->type()->pushMethod(attributeName))
	{
		lua_pushnil(pLuaState);
	}
	return 1;
}

int LuaGroupObject::newiterator(lua_State *pLuaState)
{
	t_lua_group *pObj = (t_lua_group *)lua_touserdata(pLuaState, -1);
	if(!pObj)
	{
		LuaContext::error(pLuaState,
				"Did you forget to call a record group as a function"
				" to create an iterator?");
		return 0;
	}

	lua_pushcfunction(pLuaState, iterator);
	RecordGroup::iterator i_rg;

	if(!(pObj->data().isValid()) ||
			pObj->data()->begin() == pObj->data()->end())
	{
		return 1;
	}

	lua_pushnil(pLuaState);
	t_lua_groupit *pIt = t_lua_groupit::push(pObj->context(), i_rg);
	pIt->data() = pObj->data()->begin();

	return 3;
}

int LuaGroupObject::iterator(lua_State *pLuaState)
{
	t_lua_groupit *pIt = (t_lua_groupit *)lua_touserdata(pLuaState, -1);

	if(!pIt || !(pIt->data().group().isValid()) ||
			pIt->data() == pIt->data().group()->end())
	{
//		lua_pop(pLuaState, 1);
		return 0;
	}

	t_lua_array::push(pIt->context(), *(pIt->data()));
	pIt->data()++;

	return 2;
}

void LuaGroupType::pushNewMetaTable(void)
{
	lua_State *pLuaState = m_pContext->state();

//	lua_newtable(pLuaState);
	luaL_newmetatable(pLuaState,m_mtname.c_str());

	lua_pushstring(pLuaState, "__gc");
	lua_pushcfunction(pLuaState, LuaGroupObject::destroy);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__index");
	lua_pushcfunction(pLuaState, LuaGroupObject::index);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__call");
	lua_pushcfunction(pLuaState, LuaGroupObject::newiterator);
	lua_settable(pLuaState, -3);
}


LuaGroupItType::LuaGroupItType(LuaContext *pContext,const String &mtn)
	: t_lua_groupit_t(pContext, mtn)
{
}

int LuaGroupItObject::index(lua_State *pLuaState)
{
	lua_pushnil(pLuaState);
	return 1;
}

LuaArrayType::LuaArrayType(LuaContext *pContext,const String &mtn)
	: t_lua_array_t(pContext, mtn)
{
	setMethod("check", LuaArrayObject::check);
	setMethod("length", LuaArrayObject::length);
	setMethod("add", LuaArrayObject::add);
	setMethod("remove", LuaArrayObject::remove);
}

int LuaArrayObject::index(lua_State *pLuaState)
{
	LuaArrayObject *pObj = (LuaArrayObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	if(lua_isnumber(pLuaState, -1))
	{
		int index = (int)lua_tonumber(pLuaState, -1);
		lua_pop(pLuaState, 1);
		t_lua_record::push(pObj->context(), pObj->data()->getRecord(index));
	}
	else
	{
		String attributeName = lua_tostring(pLuaState, -1);
		attributeName=pObj->aliasedName(attributeName);

		lua_pop(pLuaState, 1);

		if(!pObj->type()->pushMethod(attributeName))
		{
			lua_pushnil(pLuaState);
		}
	}
	return 1;
}

int LuaArrayObject::length(lua_State *pLuaState)
{
	LuaArrayObject *pObj = (LuaArrayObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	lua_pushnumber(pLuaState, pObj->data()->length());

	return 1;
}

int LuaArrayObject::check(lua_State *pLuaState)
{
	LuaArrayObject *pObj = (LuaArrayObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	String attributeName = lua_tostring(pLuaState, -1);
	attributeName=pObj->aliasedName(attributeName);

	lua_pushboolean(pLuaState,
			(int)pObj->data()->layout()->checkAttributeStr(attributeName));

	return 1;
}

int LuaArrayObject::add(lua_State *pLuaState)
{
	LuaArrayObject *pObj = (LuaArrayObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	LuaBaseObject *pB = (LuaBaseObject *)lua_touserdata(pLuaState, -1);
	if(!pB)
	{
		LuaContext::error(pLuaState, "invalid record");
		return 0;
	}
	if(pB->type() != pObj->type()->context()->lookup(LuaContext::e_record))
	{
		lua_pushnil(pLuaState);
		return 1;
	}

	LuaRecordObject *pR = (LuaRecordObject *)pB;

	pObj->data()->add(pR->data());

	lua_pushnil(pLuaState);
	return 1;
}

int LuaArrayObject::remove(lua_State *pLuaState)
{
	LuaArrayObject *pObj = (LuaArrayObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	LuaBaseObject *pB = (LuaBaseObject *)lua_touserdata(pLuaState, -1);
	if(!pB)
	{
		LuaContext::error(pLuaState, "invalid record");
		return 0;
	}
	if(pB->type() != pObj->type()->context()->lookup(LuaContext::e_record))
	{
		lua_pushnil(pLuaState);
		return 1;
	}

	LuaRecordObject *pR = (LuaRecordObject *)pB;

	pObj->data()->remove(pR->data());

	lua_pushnil(pLuaState);
	return 1;
}

int LuaArrayObject::newiterator(lua_State *pLuaState)
{
	t_lua_array *pObj = (t_lua_array *)lua_touserdata(pLuaState, -1);
	if(!pObj)
	{
		LuaContext::error(pLuaState,
				"Did you forget to call a record array as a function"
				" to create an iterator?");
		return 0;
	}

	lua_pushcfunction(pLuaState, iterator);
	LuaArrayContext arrayContext;

	if(!(pObj->data().isValid()) || pObj->data()->length() == 0)
	{
		return 1;
	}

	lua_pushnil(pLuaState);
	t_lua_arrayit *pIt = t_lua_arrayit::push(pObj->context(), arrayContext);
	pIt->data().m_index = 0;
	pIt->data().m_spRecordArray=pObj->data();

	return 3;
}

int LuaArrayObject::iterator(lua_State *pLuaState)
{
	t_lua_arrayit *pIt = (t_lua_arrayit *)lua_touserdata(pLuaState, -1);

	if(!pIt || pIt->data().m_spRecordArray.isNull() ||
			pIt->data().m_index>=pIt->data().m_spRecordArray->length())
	{
//		lua_pop(pLuaState, 1);
		return 0;
	}

	t_lua_record::push(pIt->context(),
			pIt->data().m_spRecordArray->getRecord(pIt->data().m_index));
	pIt->data().m_index++;

	return 2;
}

void LuaArrayType::pushNewMetaTable(void)
{
	lua_State *pLuaState = m_pContext->state();

//	lua_newtable(pLuaState);
	luaL_newmetatable(pLuaState,m_mtname.c_str());

	lua_pushstring(pLuaState, "__gc");
	lua_pushcfunction(pLuaState, LuaArrayObject::destroy);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__index");
	lua_pushcfunction(pLuaState, LuaArrayObject::index);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__call");
	lua_pushcfunction(pLuaState, LuaArrayObject::newiterator);
	lua_settable(pLuaState, -3);
}

LuaArrayItType::LuaArrayItType(LuaContext *pContext,const String &mtn)
	: t_lua_arrayit_t(pContext, mtn)
{
}

int LuaArrayItObject::index(lua_State *pLuaState)
{
	lua_pushnil(pLuaState);
	return 1;
}

LuaIntArrayType::LuaIntArrayType(LuaContext *pContext,const String &mtn)
	: t_lua_intarray_t(pContext, mtn)
{
	setMethod("length", LuaIntArrayObject::length);
	setMethod("add", LuaIntArrayObject::add);
	setMethod("resize", LuaIntArrayObject::resize);
}

int LuaIntArrayObject::index(lua_State *pLuaState)
{
	LuaIntArrayObject *pObj = (LuaIntArrayObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	if(lua_isnumber(pLuaState, -1))
	{
		int index = (int)lua_tonumber(pLuaState, -1);
		lua_pop(pLuaState, 1);
		lua_pushnumber(pLuaState, (*pObj->data())[index]);
		return 1;
	}

	String methodName = lua_tostring(pLuaState, -1);

	lua_pop(pLuaState, 1);

	if(!pObj->type()->pushMethod(methodName))
	{
		return 0;
	}

	return 1;
}

int LuaIntArrayObject::length(lua_State *pLuaState)
{
	LuaIntArrayObject *pObj = (LuaIntArrayObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	lua_pushnumber(pLuaState, (*pObj->data()).size());
	return 1;
}

int LuaIntArrayObject::add(lua_State *pLuaState)
{
	LuaIntArrayObject *pObj = (LuaIntArrayObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	if(!lua_isnumber(pLuaState, -1))
	{
		LuaContext::error(pLuaState, "bad value type");
		return 0;
	}

	const int newValue = (int)lua_tonumber(pLuaState, -1);
	lua_pop(pLuaState, 1);

	const I32 newIndex=pObj->data()->size();
	pObj->data()->push_back(newValue);

	lua_pushnumber(pLuaState, newIndex);
	return 1;
}

int LuaIntArrayObject::resize(lua_State *pLuaState)
{
	LuaIntArrayObject *pObj = (LuaIntArrayObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	if(!lua_isnumber(pLuaState, -1))
	{
		LuaContext::error(pLuaState, "bad size type");
		return 0;
	}

	const int newSize = (int)lua_tonumber(pLuaState, -1);
	lua_pop(pLuaState, 1);

	pObj->data()->resize(newSize);

	lua_pushnumber(pLuaState, newSize);
	return 1;
}

int LuaIntArrayObject::newiterator(lua_State *pLuaState)
{
	t_lua_intarray *pObj = (t_lua_intarray *)lua_touserdata(pLuaState, -1);
	if(!pObj)
	{
		LuaContext::error(pLuaState,
				"Did you forget to call a int array as a function"
				" to create an iterator?");
		return 0;
	}

	lua_pushcfunction(pLuaState, iterator);
	LuaIntArrayContext arrayContext;

	if(pObj->data()->size() == 0)
	{
		return 1;
	}

	lua_pushnil(pLuaState);
	t_lua_intarrayit *pIt =
			t_lua_intarrayit::push(pObj->context(), arrayContext);
	pIt->data().m_pIntArray=pObj->data();
	pIt->data().m_index = 0;

	return 3;
}

int LuaIntArrayObject::iterator(lua_State *pLuaState)
{
	t_lua_intarrayit *pIt = (t_lua_intarrayit *)lua_touserdata(pLuaState, -1);

	if(!pIt || pIt->data().m_index>=I32(pIt->data().m_pIntArray->size()))
	{
//		lua_pop(pLuaState, 1);
		return 0;
	}

	t_lua_record::push(pIt->context(),
			(*pIt->data().m_pIntArray)[pIt->data().m_index]);
	pIt->data().m_index++;

	return 2;
}

void LuaIntArrayType::pushNewMetaTable(void)
{
	lua_State *pLuaState = m_pContext->state();

//	lua_newtable(pLuaState);
	luaL_newmetatable(pLuaState,m_mtname.c_str());

	lua_pushstring(pLuaState, "__gc");
	lua_pushcfunction(pLuaState, LuaIntArrayObject::destroy);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__index");
	lua_pushcfunction(pLuaState, LuaIntArrayObject::index);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__call");
	lua_pushcfunction(pLuaState, LuaIntArrayObject::newiterator);
	lua_settable(pLuaState, -3);
}

LuaIntArrayItType::LuaIntArrayItType(LuaContext *pContext,const String &mtn)
	: t_lua_intarrayit_t(pContext, mtn)
{
}

int LuaIntArrayItObject::index(lua_State *pLuaState)
{
	lua_pushnil(pLuaState);
	return 1;
}

LuaVectorType::LuaVectorType(LuaContext *pContext,const String &mtn)
	: t_lua_vector_t(pContext, mtn)
{
}

int LuaVectorObject::index(lua_State *pLuaState)
{
	LuaVectorObject *pObj = (LuaVectorObject *)selfIsFirst(pLuaState);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	if(lua_isnumber(pLuaState, -1))
	{
		int index = (int)lua_tonumber(pLuaState, -1);
		lua_pop(pLuaState, 1);
		lua_pushnumber(pLuaState, pObj->data()[index]);
	}
	else
	{
		String methodName = lua_tostring(pLuaState, -1);

		lua_pop(pLuaState, 1);

		if(!pObj->type()->pushMethod(methodName))
		{
			lua_pushnil(pLuaState);
		}
	}
	return 1;
}

int LuaVectorObject::newindex(lua_State *pLuaState)
{
	LuaVectorObject *pObj = (LuaVectorObject *)lua_touserdata(pLuaState, 1);
	if(!pObj) { lua_pushnil(pLuaState); return 1; }

	if(lua_isnumber(pLuaState, -2))
	{
		int index = (int)lua_tonumber(pLuaState, -2);
		float value = (float)lua_tonumber(pLuaState, -1);

		pObj->data()[index] = value;

		lua_pushnumber(pLuaState, value);
	}
	else
	{
		lua_pushnil(pLuaState);
	}

	return 1;
}

void LuaVectorType::pushNewMetaTable(void)
{
	lua_State *pLuaState = m_pContext->state();

//	lua_newtable(pLuaState);
	luaL_newmetatable(pLuaState,m_mtname.c_str());

	lua_pushstring(pLuaState, "__gc");
	lua_pushcfunction(pLuaState, LuaVectorObject::destroy);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__index");
	lua_pushcfunction(pLuaState, LuaVectorObject::index);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__newindex");
	lua_pushcfunction(pLuaState, LuaVectorObject::newindex);
	lua_settable(pLuaState, -3);
}

LuaVector4Type::LuaVector4Type(LuaContext *pContext,const String &mtn)
	: t_lua_vector4_t(pContext, mtn)
{
	//* unit normalize magnitude dot cross3 set
	setMethod("unit",		LuaVector4Object::unit);
	setMethod("normalize",	LuaVector4Object::normalize);
	setMethod("magnitude",	LuaVector4Object::magnitude);
	setMethod("dot",		LuaVector4Object::dot);
	setMethod("cross3",		LuaVector4Object::cross3);
	setMethod("set",		LuaVector4Object::set);
}

int LuaVector4Object::index(lua_State *pLuaState)
{
	LuaVector4Object *pObj = (LuaVector4Object *)selfIsFirst(pLuaState);
	if(!pObj)
	{
		return 0;
	}

	if(lua_isnumber(pLuaState, -1))
	{
		int index = (int)lua_tonumber(pLuaState, -1);
		lua_pop(pLuaState, 1);
		lua_pushnumber(pLuaState, (*pObj->data())[index]);
		return 1;
	}

	String methodName = lua_tostring(pLuaState, -1);

	lua_pop(pLuaState, 1);

	if(!pObj->type()->pushMethod(methodName))
	{
		return 0;
	}

	return 1;
}

int LuaVector4Object::newindex(lua_State *pLuaState)
{
	LuaVector4Object *pObj = (LuaVector4Object *)lua_touserdata(pLuaState, 1);
	if(!pObj|| !lua_isnumber(pLuaState, -2))
	{
		return 0;
	}

	int index = (int)lua_tonumber(pLuaState, -2);
	float value = (float)lua_tonumber(pLuaState, -1);

	setAt(*pObj->data(),index,(Real)value);

	lua_pushnumber(pLuaState, value);
	return 1;
}

int LuaVector4Object::eq(lua_State *pLuaState)
{
	LuaVector4Object *pObj = (LuaVector4Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector4Object* pOther;
	if((pOther=reinterpret_cast<LuaVector4Object*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		lua_pushboolean(pLuaState,*pObj->data() == *pOther->data());
		return 1;
	}

	feLog("LuaVector4Object::eq incompatible argType \"%s\"\n",
			argType(pLuaState,2).c_str());
	return 0;
}

//* unary negation
int LuaVector4Object::unm(lua_State *pLuaState)
{
	LuaVector4Object *pObj = (LuaVector4Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	Vector4* pNegative=new Vector4(-(*pObj->data()));
	t_lua_vector4::push(pObj->context(),pNegative);
	return 1;
}

int LuaVector4Object::add(lua_State *pLuaState)
{
	LuaVector4Object *pObj = (LuaVector4Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector4Object* pOther;
	if((pOther=reinterpret_cast<LuaVector4Object*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		Vector4* pSum=new Vector4(*pObj->data() + *pOther->data());
		t_lua_vector4::push(pObj->context(),pSum);
		return 1;
	}

	feLog("LuaVector4Object::add incompatible argType \"%s\"\n",
			argType(pLuaState,2).c_str());
	return 0;
}

int LuaVector4Object::sub(lua_State *pLuaState)
{
	LuaVector4Object *pObj = (LuaVector4Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector4Object* pOther;
	if((pOther=reinterpret_cast<LuaVector4Object*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		Vector4* pDiff=new Vector4(*pObj->data() - *pOther->data());
		t_lua_vector4::push(pObj->context(),pDiff);
		return 1;
	}

	feLog("LuaVector4Object::sub incompatible argType \"%s\"\n",
			argType(pLuaState,2).c_str());
	return 0;
}

int LuaVector4Object::mul(lua_State *pLuaState)
{
	LuaVector4Object *pObj = (LuaVector4Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	F32 scalar=0.0f;
	if(lua_isnumber(pLuaState, -1))
	{
		scalar = (F32)lua_tonumber(pLuaState, -1);
	}
	else if(lua_isnumber(pLuaState, -2))
	{
		scalar = (F32)lua_tonumber(pLuaState, -2);
	}
	else
	{
		return 0;
	}

	Vector4* pSum=new Vector4(*pObj->data() * (Real)scalar);
	t_lua_vector4::push(pObj->context(),pSum);
	return 1;
}

int LuaVector4Object::unit(lua_State *pLuaState)
{
	LuaVector4Object *pObj = (LuaVector4Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	Vector4* pNormal=new Vector4(fe::unit(*pObj->data()));
	t_lua_vector4::push(pObj->context(),pNormal);
	return 1;
}

int LuaVector4Object::normalize(lua_State *pLuaState)
{
	LuaVector4Object *pObj = (LuaVector4Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	fe::normalize(*pObj->data());
	t_lua_vector4::push(pObj->context(),pObj->data());
	return 1;
}

int LuaVector4Object::magnitude(lua_State *pLuaState)
{
	LuaVector4Object *pObj = (LuaVector4Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	F32 length=fe::magnitude(*pObj->data());
	lua_pushnumber(pLuaState, length);
	return 1;
}

int LuaVector4Object::dot(lua_State *pLuaState)
{
	LuaVector4Object *pObj = (LuaVector4Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector4Object* pOther;
	if((pOther=reinterpret_cast<LuaVector4Object*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		F32 dot=fe::dot(*pObj->data(),*pOther->data());
		lua_pushnumber(pLuaState, dot);
		return 1;
	}

	feLog("LuaVector4Object::dot incompatible argType \"%s\"\n",
			argType(pLuaState,2).c_str());
	return 0;
}

int LuaVector4Object::cross3(lua_State *pLuaState)
{
	LuaVector4Object *pObj = (LuaVector4Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector4Object* pOther;
	if((pOther=reinterpret_cast<LuaVector4Object*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		Vector4* pCross=
				new Vector4(fe::cross3(*pObj->data(),*pOther->data()));
		t_lua_vector4::push(pObj->context(),pCross);
		return 1;
	}

	feLog("LuaVector4Object::cross3 incompatible argType \"%s\"\n",
			argType(pLuaState,2).c_str());
	return 0;
}

int LuaVector4Object::set(lua_State *pLuaState)
{
	LuaVector4Object *pObj = (LuaVector4Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector4Object* pOther;
	if((pOther=reinterpret_cast<LuaVector4Object*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		*pObj->data()=*pOther->data();
		t_lua_vector4::push(pObj->context(),pObj->data());
		return 1;
	}

	Real vector[4]={ 0.0f,0.0f,0.0f,0.0f };
	if(lua_isnumber(pLuaState, 2))
	{
		vector[0]=(Real)lua_tonumber(pLuaState, 2);
	}
	if(lua_isnumber(pLuaState, 3))
	{
		vector[1]=(Real)lua_tonumber(pLuaState, 3);
	}
	if(lua_isnumber(pLuaState, 4))
	{
		vector[2]=(Real)lua_tonumber(pLuaState, 4);
	}
	if(lua_isnumber(pLuaState, 5))
	{
		vector[3]=(Real)lua_tonumber(pLuaState, 5);
	}

	fe::set(*pObj->data(),vector[0],vector[1],vector[2],vector[3]);
	t_lua_vector4::push(pObj->context(),pObj->data());
	return 1;
}

void LuaVector4Type::pushNewMetaTable(void)
{
	lua_State *pLuaState = m_pContext->state();

//	lua_newtable(pLuaState);
	luaL_newmetatable(pLuaState,m_mtname.c_str());

	// call index newindex add sub mul div unm pow concat lt le eq

	lua_pushstring(pLuaState, "__gc");
	lua_pushcfunction(pLuaState, LuaVector4Object::destroy);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__index");
	lua_pushcfunction(pLuaState, LuaVector4Object::index);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__newindex");
	lua_pushcfunction(pLuaState, LuaVector4Object::newindex);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__eq");
	lua_pushcfunction(pLuaState, LuaVector4Object::eq);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__unm");
	lua_pushcfunction(pLuaState, LuaVector4Object::unm);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__add");
	lua_pushcfunction(pLuaState, LuaVector4Object::add);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__sub");
	lua_pushcfunction(pLuaState, LuaVector4Object::sub);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__mul");
	lua_pushcfunction(pLuaState, LuaVector4Object::mul);
	lua_settable(pLuaState, -3);
}


// ------------------------------------------------------------------
// ------------------------------------------------------------------
// ------------------------------------------------------------------


LuaVector4fType::LuaVector4fType(LuaContext *pContext,const String &mtn)
	: t_lua_vector4_t(pContext, mtn)
{
	//* unit normalize magnitude dot cross3 set
	setMethod("unit",		LuaVector4fObject::unit);
	setMethod("normalize",	LuaVector4fObject::normalize);
	setMethod("magnitude",	LuaVector4fObject::magnitude);
	setMethod("dot",		LuaVector4fObject::dot);
	setMethod("cross3",		LuaVector4fObject::cross3);
	setMethod("set",		LuaVector4fObject::set);
}

int LuaVector4fObject::index(lua_State *pLuaState)
{
	LuaVector4fObject *pObj = (LuaVector4fObject *)selfIsFirst(pLuaState);
	if(!pObj)
	{
		return 0;
	}

	if(lua_isnumber(pLuaState, -1))
	{
		int index = (int)lua_tonumber(pLuaState, -1);
		lua_pop(pLuaState, 1);
		lua_pushnumber(pLuaState, (*pObj->data())[index]);
		return 1;
	}

	String methodName = lua_tostring(pLuaState, -1);

	lua_pop(pLuaState, 1);

	if(!pObj->type()->pushMethod(methodName))
	{
		return 0;
	}

	return 1;
}

int LuaVector4fObject::newindex(lua_State *pLuaState)
{
	LuaVector4fObject *pObj = (LuaVector4fObject *)lua_touserdata(pLuaState, 1);
	if(!pObj|| !lua_isnumber(pLuaState, -2))
	{
		return 0;
	}

	int index = (int)lua_tonumber(pLuaState, -2);
	float value = (float)lua_tonumber(pLuaState, -1);

	setAt(*pObj->data(),index,value);

	lua_pushnumber(pLuaState, value);
	return 1;
}

int LuaVector4fObject::eq(lua_State *pLuaState)
{
	LuaVector4fObject *pObj = (LuaVector4fObject *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector4fObject* pOther;
	if((pOther=reinterpret_cast<LuaVector4fObject*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		lua_pushboolean(pLuaState,*pObj->data() == *pOther->data());
		return 1;
	}

	feLog("LuaVector4fObject::eq incompatible argType \"%s\"\n",
			argType(pLuaState,2).c_str());
	return 0;
}

//* unary negation
int LuaVector4fObject::unm(lua_State *pLuaState)
{
	LuaVector4fObject *pObj = (LuaVector4fObject *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	Vector4f* pNegative=new Vector4f(-(*pObj->data()));
	t_lua_vector4f::push(pObj->context(),pNegative);
	return 1;
}

int LuaVector4fObject::add(lua_State *pLuaState)
{
	LuaVector4fObject *pObj = (LuaVector4fObject *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector4fObject* pOther;
	if((pOther=reinterpret_cast<LuaVector4fObject*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		Vector4f* pSum=new Vector4f(*pObj->data() + *pOther->data());
		t_lua_vector4f::push(pObj->context(),pSum);
		return 1;
	}

	feLog("LuaVector4fObject::add incompatible argType \"%s\"\n",
			argType(pLuaState,2).c_str());
	return 0;
}

int LuaVector4fObject::sub(lua_State *pLuaState)
{
	LuaVector4fObject *pObj = (LuaVector4fObject *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector4fObject* pOther;
	if((pOther=reinterpret_cast<LuaVector4fObject*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		Vector4f* pDiff=new Vector4f(*pObj->data() - *pOther->data());
		t_lua_vector4f::push(pObj->context(),pDiff);
		return 1;
	}

	feLog("LuaVector4fObject::sub incompatible argType \"%s\"\n",
			argType(pLuaState,2).c_str());
	return 0;
}

int LuaVector4fObject::mul(lua_State *pLuaState)
{
	LuaVector4fObject *pObj = (LuaVector4fObject *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	F32 scalar=0.0f;
	if(lua_isnumber(pLuaState, -1))
	{
		scalar = (F32)lua_tonumber(pLuaState, -1);
	}
	else if(lua_isnumber(pLuaState, -2))
	{
		scalar = (F32)lua_tonumber(pLuaState, -2);
	}
	else
	{
		return 0;
	}

	Vector4f* pSum=new Vector4f(*pObj->data() * scalar);
	t_lua_vector4f::push(pObj->context(),pSum);
	return 1;
}

int LuaVector4fObject::unit(lua_State *pLuaState)
{
	LuaVector4fObject *pObj = (LuaVector4fObject *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	Vector4f* pNormal=new Vector4f(fe::unit(*pObj->data()));
	t_lua_vector4f::push(pObj->context(),pNormal);
	return 1;
}

int LuaVector4fObject::normalize(lua_State *pLuaState)
{
	LuaVector4fObject *pObj = (LuaVector4fObject *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	fe::normalize(*pObj->data());
	t_lua_vector4f::push(pObj->context(),pObj->data());
	return 1;
}

int LuaVector4fObject::magnitude(lua_State *pLuaState)
{
	LuaVector4fObject *pObj = (LuaVector4fObject *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	F32 length=fe::magnitude(*pObj->data());
	lua_pushnumber(pLuaState, length);
	return 1;
}

int LuaVector4fObject::dot(lua_State *pLuaState)
{
	LuaVector4fObject *pObj = (LuaVector4fObject *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector4fObject* pOther;
	if((pOther=reinterpret_cast<LuaVector4fObject*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		F32 dot=fe::dot(*pObj->data(),*pOther->data());
		lua_pushnumber(pLuaState, dot);
		return 1;
	}

	feLog("LuaVector4fObject::dot incompatible argType \"%s\"\n",
			argType(pLuaState,2).c_str());
	return 0;
}

int LuaVector4fObject::cross3(lua_State *pLuaState)
{
	LuaVector4fObject *pObj = (LuaVector4fObject *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector4fObject* pOther;
	if((pOther=reinterpret_cast<LuaVector4fObject*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		Vector4f* pCross=
				new Vector4f(fe::cross3(*pObj->data(),*pOther->data()));
		t_lua_vector4f::push(pObj->context(),pCross);
		return 1;
	}

	feLog("LuaVector4fObject::cross3 incompatible argType \"%s\"\n",
			argType(pLuaState,2).c_str());
	return 0;
}

int LuaVector4fObject::set(lua_State *pLuaState)
{
	LuaVector4fObject *pObj = (LuaVector4fObject *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector4fObject* pOther;
	if((pOther=reinterpret_cast<LuaVector4fObject*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		*pObj->data()=*pOther->data();
		t_lua_vector4f::push(pObj->context(),pObj->data());
		return 1;
	}

	F32 vector[4]={ 0.0f,0.0f,0.0f,0.0f };
	if(lua_isnumber(pLuaState, 2))
	{
		vector[0]=(Real)lua_tonumber(pLuaState, 2);
	}
	if(lua_isnumber(pLuaState, 3))
	{
		vector[1]=(Real)lua_tonumber(pLuaState, 3);
	}
	if(lua_isnumber(pLuaState, 4))
	{
		vector[2]=(Real)lua_tonumber(pLuaState, 4);
	}
	if(lua_isnumber(pLuaState, 5))
	{
		vector[3]=(Real)lua_tonumber(pLuaState, 5);
	}

	fe::set(*pObj->data(),vector[0],vector[1],vector[2],vector[3]);
	t_lua_vector4f::push(pObj->context(),pObj->data());
	return 1;
}

void LuaVector4fType::pushNewMetaTable(void)
{
	lua_State *pLuaState = m_pContext->state();

//	lua_newtable(pLuaState);
	luaL_newmetatable(pLuaState,m_mtname.c_str());

	// call index newindex add sub mul div unm pow concat lt le eq

	lua_pushstring(pLuaState, "__gc");
	lua_pushcfunction(pLuaState, LuaVector4fObject::destroy);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__index");
	lua_pushcfunction(pLuaState, LuaVector4fObject::index);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__newindex");
	lua_pushcfunction(pLuaState, LuaVector4fObject::newindex);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__eq");
	lua_pushcfunction(pLuaState, LuaVector4fObject::eq);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__unm");
	lua_pushcfunction(pLuaState, LuaVector4fObject::unm);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__add");
	lua_pushcfunction(pLuaState, LuaVector4fObject::add);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__sub");
	lua_pushcfunction(pLuaState, LuaVector4fObject::sub);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__mul");
	lua_pushcfunction(pLuaState, LuaVector4fObject::mul);
	lua_settable(pLuaState, -3);
}


// ------------------------------------------------------------------
// ------------------------------------------------------------------
// ------------------------------------------------------------------

LuaVector3Type::LuaVector3Type(LuaContext *pContext,const String &mtn)
	: t_lua_vector3_t(pContext, mtn)
{
	//* unit normalize magnitude dot cross3 set
	setMethod("unit",		LuaVector3Object::unit);
	setMethod("normalize",	LuaVector3Object::normalize);
	setMethod("magnitude",	LuaVector3Object::magnitude);
	setMethod("dot",		LuaVector3Object::dot);
	setMethod("cross3",		LuaVector3Object::cross3);
	setMethod("set",		LuaVector3Object::set);
}

int LuaVector3Object::index(lua_State *pLuaState)
{
	LuaVector3Object *pObj = (LuaVector3Object *)selfIsFirst(pLuaState);
	if(!pObj)
	{
		return 0;
	}

	if(lua_isnumber(pLuaState, -1))
	{
		int index = (int)lua_tonumber(pLuaState, -1);
		lua_pop(pLuaState, 1);
		lua_pushnumber(pLuaState, (*pObj->data())[index]);
		return 1;
	}

	String methodName = lua_tostring(pLuaState, -1);

	lua_pop(pLuaState, 1);

	if(!pObj->type()->pushMethod(methodName))
	{
		return 0;
	}

	return 1;
}

int LuaVector3Object::newindex(lua_State *pLuaState)
{
	LuaVector3Object *pObj = (LuaVector3Object *)lua_touserdata(pLuaState, 1);
	if(!pObj|| !lua_isnumber(pLuaState, -2))
	{
		return 0;
	}

	int index = (int)lua_tonumber(pLuaState, -2);
	float value = (float)lua_tonumber(pLuaState, -1);

	setAt(*pObj->data(),index,(Real)value);

	lua_pushnumber(pLuaState, value);
	return 1;
}

int LuaVector3Object::eq(lua_State *pLuaState)
{
	LuaVector3Object *pObj = (LuaVector3Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector3Object* pOther;
	if((pOther=reinterpret_cast<LuaVector3Object*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		lua_pushboolean(pLuaState,*pObj->data() == *pOther->data());
		return 1;
	}

	feLog("LuaVector3Object::eq incompatible argType \"%s\"\n",
			argType(pLuaState,2).c_str());
	return 0;
}

//* unary negation
int LuaVector3Object::unm(lua_State *pLuaState)
{
	LuaVector3Object *pObj = (LuaVector3Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	Vector3* pNegative=new Vector3(-(*pObj->data()));
	t_lua_vector3::push(pObj->context(),pNegative);
	return 1;
}

int LuaVector3Object::add(lua_State *pLuaState)
{
	LuaVector3Object *pObj = (LuaVector3Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector3Object* pOther;
	if((pOther=reinterpret_cast<LuaVector3Object*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		Vector3* pSum=new Vector3(*pObj->data() + *pOther->data());
		t_lua_vector3::push(pObj->context(),pSum);
		return 1;
	}

	feLog("LuaVector3Object::add incompatible argType \"%s\"\n",
			argType(pLuaState,2).c_str());
	return 0;
}

int LuaVector3Object::sub(lua_State *pLuaState)
{
	LuaVector3Object *pObj = (LuaVector3Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector3Object* pOther;
	if((pOther=reinterpret_cast<LuaVector3Object*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		Vector3 *pDiff=new Vector3(*pObj->data() - *pOther->data());
		t_lua_vector3::push(pObj->context(),pDiff);
		return 1;
	}

	feLog("LuaVector3Object::sub incompatible argType \"%s\"\n",
			argType(pLuaState,2).c_str());
	return 0;
}

int LuaVector3Object::mul(lua_State *pLuaState)
{
	LuaVector3Object *pObj = (LuaVector3Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	F32 scalar=0.0f;
	if(lua_isnumber(pLuaState, -1))
	{
		scalar = (F32)lua_tonumber(pLuaState, -1);
	}
	else if(lua_isnumber(pLuaState, -2))
	{
		scalar = (F32)lua_tonumber(pLuaState, -2);
	}
	else
	{
		return 0;
	}

	Vector3* pSum=new Vector3(*pObj->data() * (Real)scalar);
	t_lua_vector3::push(pObj->context(),pSum);
	return 1;
}

int LuaVector3Object::tostring(lua_State *pLuaState)
{
	LuaVector3Object *pObj = (LuaVector3Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	Vector3& v= *(pObj->data());
	lua_pushstring(pLuaState,c_print(v));

	return 1;
}

int LuaVector3Object::unit(lua_State *pLuaState)
{
	LuaVector3Object *pObj = (LuaVector3Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	Vector3* pNormal=new Vector3(fe::unit(*pObj->data()));
	t_lua_vector3::push(pObj->context(),pNormal);
	return 1;
}

int LuaVector3Object::normalize(lua_State *pLuaState)
{
	LuaVector3Object *pObj = (LuaVector3Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	fe::normalize(*pObj->data());
	t_lua_vector3::push(pObj->context(),pObj->data());
	return 1;
}

int LuaVector3Object::magnitude(lua_State *pLuaState)
{
	LuaVector3Object *pObj = (LuaVector3Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	F32 length=fe::magnitude(*pObj->data());
	lua_pushnumber(pLuaState, length);
	return 1;
}

int LuaVector3Object::dot(lua_State *pLuaState)
{
	LuaVector3Object *pObj = (LuaVector3Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector3Object* pOther;
	if((pOther=reinterpret_cast<LuaVector3Object*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		F32 dot=fe::dot(*pObj->data(),*pOther->data());
		lua_pushnumber(pLuaState, dot);
		return 1;
	}

	feLog("LuaVector3Object::dot incompatible argType \"%s\"\n",
			argType(pLuaState,2).c_str());
	return 0;
}

int LuaVector3Object::cross3(lua_State *pLuaState)
{
	LuaVector3Object *pObj = (LuaVector3Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector3Object* pOther;
	if((pOther=reinterpret_cast<LuaVector3Object*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		Vector3* pCross=
				new Vector3(fe::cross3(*pObj->data(),*pOther->data()));
		t_lua_vector3::push(pObj->context(),pCross);
		return 1;
	}

	feLog("LuaVector3Object::cross3 incompatible argType \"%s\"\n",
			argType(pLuaState,2).c_str());
	return 0;
}

int LuaVector3Object::set(lua_State *pLuaState)
{
	LuaVector3Object *pObj = (LuaVector3Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector3Object* pOther;
	if((pOther=reinterpret_cast<LuaVector3Object*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		*pObj->data()=*pOther->data();
		t_lua_vector3::push(pObj->context(),pObj->data());
		return 1;
	}

	Real vector[3]={ 0.0f,0.0f,0.0f };
	if(lua_isnumber(pLuaState, 2))
	{
		vector[0]=(Real)lua_tonumber(pLuaState, 2);
	}
	if(lua_isnumber(pLuaState, 3))
	{
		vector[1]=(Real)lua_tonumber(pLuaState, 3);
	}
	if(lua_isnumber(pLuaState, 4))
	{
		vector[2]=(Real)lua_tonumber(pLuaState, 4);
	}

	fe::set<Real>(*pObj->data(),vector[0],vector[1],vector[2]);
	t_lua_vector3::push(pObj->context(),pObj->data());
	return 1;
}

void LuaVector3Type::pushNewMetaTable(void)
{
	lua_State *pLuaState = m_pContext->state();

//	lua_newtable(pLuaState);
	luaL_newmetatable(pLuaState,m_mtname.c_str());

	// call index newindex add sub mul div unm pow concat lt le eq

	lua_pushstring(pLuaState, "__gc");
	lua_pushcfunction(pLuaState, LuaVector3Object::destroy);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__index");
	lua_pushcfunction(pLuaState, LuaVector3Object::index);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__newindex");
	lua_pushcfunction(pLuaState, LuaVector3Object::newindex);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__eq");
	lua_pushcfunction(pLuaState, LuaVector3Object::eq);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__unm");
	lua_pushcfunction(pLuaState, LuaVector3Object::unm);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__add");
	lua_pushcfunction(pLuaState, LuaVector3Object::add);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__sub");
	lua_pushcfunction(pLuaState, LuaVector3Object::sub);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__mul");
	lua_pushcfunction(pLuaState, LuaVector3Object::mul);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__tostring");
	lua_pushcfunction(pLuaState, LuaVector3Object::tostring);
	lua_settable(pLuaState, -3);
}

// ------------------------------------------------------------------
// ------------------------------------------------------------------
// ------------------------------------------------------------------

LuaVector2Type::LuaVector2Type(LuaContext *pContext,const String &mtn)
	: t_lua_vector2_t(pContext, mtn)
{
	setMethod("unit",		LuaVector2Object::unit);
	setMethod("normalize",	LuaVector2Object::normalize);
	setMethod("magnitude",	LuaVector2Object::magnitude);
	setMethod("dot",		LuaVector2Object::dot);
	setMethod("set",		LuaVector2Object::set);
}

int LuaVector2Object::index(lua_State *pLuaState)
{
	LuaVector2Object *pObj = (LuaVector2Object *)selfIsFirst(pLuaState);
	if(!pObj)
	{
		return 0;
	}

	if(lua_isnumber(pLuaState, -1))
	{
		int index = (int)lua_tonumber(pLuaState, -1);
		lua_pop(pLuaState, 1);
		lua_pushnumber(pLuaState, (*pObj->data())[index]);
		return 1;
	}

	String methodName = lua_tostring(pLuaState, -1);

	lua_pop(pLuaState, 1);

	if(!pObj->type()->pushMethod(methodName))
	{
		return 0;
	}

	return 1;
}

int LuaVector2Object::newindex(lua_State *pLuaState)
{
	LuaVector2Object *pObj = (LuaVector2Object *)lua_touserdata(pLuaState, 1);
	if(!pObj|| !lua_isnumber(pLuaState, -2))
	{
		return 0;
	}

	int index = (int)lua_tonumber(pLuaState, -2);
	float value = (float)lua_tonumber(pLuaState, -1);

	setAt(*pObj->data(),index,(Real)value);

	lua_pushnumber(pLuaState, value);
	return 1;
}

int LuaVector2Object::eq(lua_State *pLuaState)
{
	LuaVector2Object *pObj = (LuaVector2Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector2Object* pOther;
	if((pOther=reinterpret_cast<LuaVector2Object*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		lua_pushboolean(pLuaState,*pObj->data() == *pOther->data());
		return 1;
	}

	feLog("LuaVector2Object::eq incompatible argType \"%s\"\n",
			argType(pLuaState,2).c_str());
	return 0;
}

//* unary negation
int LuaVector2Object::unm(lua_State *pLuaState)
{
	LuaVector2Object *pObj = (LuaVector2Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	Vector2* pNegative=new Vector2(-(*pObj->data()));
	t_lua_vector2::push(pObj->context(),pNegative);
	return 1;
}

int LuaVector2Object::add(lua_State *pLuaState)
{
	LuaVector2Object *pObj = (LuaVector2Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector2Object* pOther;
	if((pOther=reinterpret_cast<LuaVector2Object*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		Vector2* pSum=new Vector2(*pObj->data() + *pOther->data());
		t_lua_vector2::push(pObj->context(),pSum);
		return 1;
	}

	feLog("LuaVector3Object::add incompatible argType \"%s\"\n",
			argType(pLuaState,2).c_str());
	return 0;
}

int LuaVector2Object::sub(lua_State *pLuaState)
{
	LuaVector2Object *pObj = (LuaVector2Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector2Object* pOther;
	if((pOther=reinterpret_cast<LuaVector2Object*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		Vector2 *pDiff=new Vector2(*pObj->data() - *pOther->data());
		t_lua_vector2::push(pObj->context(),pDiff);
		return 1;
	}

	feLog("LuaVector2Object::sub incompatible argType \"%s\"\n",
			argType(pLuaState,2).c_str());
	return 0;
}

int LuaVector2Object::mul(lua_State *pLuaState)
{
	LuaVector2Object *pObj = (LuaVector2Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	F32 scalar=0.0f;
	if(lua_isnumber(pLuaState, -1))
	{
		scalar = (F32)lua_tonumber(pLuaState, -1);
	}
	else if(lua_isnumber(pLuaState, -2))
	{
		scalar = (F32)lua_tonumber(pLuaState, -2);
	}
	else
	{
		return 0;
	}

	Vector2* pSum=new Vector2(*pObj->data() * (Real)scalar);
	t_lua_vector2::push(pObj->context(),pSum);
	return 1;
}

int LuaVector2Object::unit(lua_State *pLuaState)
{
	LuaVector2Object *pObj = (LuaVector2Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	Vector2* pNormal=new Vector2(fe::unit(*pObj->data()));
	t_lua_vector2::push(pObj->context(),pNormal);
	return 1;
}

int LuaVector2Object::normalize(lua_State *pLuaState)
{
	LuaVector2Object *pObj = (LuaVector2Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	fe::normalize(*pObj->data());
	t_lua_vector2::push(pObj->context(),pObj->data());
	return 1;
}

int LuaVector2Object::magnitude(lua_State *pLuaState)
{
	LuaVector2Object *pObj = (LuaVector2Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	F32 length=fe::magnitude(*pObj->data());
	lua_pushnumber(pLuaState, length);
	return 1;
}

int LuaVector2Object::dot(lua_State *pLuaState)
{
	LuaVector2Object *pObj = (LuaVector2Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector2Object* pOther;
	if((pOther=reinterpret_cast<LuaVector2Object*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		F32 dot=fe::dot(*pObj->data(),*pOther->data());
		lua_pushnumber(pLuaState, dot);
		return 1;
	}

	feLog("LuaVector2Object::dot incompatible argType \"%s\"\n",
			argType(pLuaState,2).c_str());
	return 0;
}


int LuaVector2Object::set(lua_State *pLuaState)
{
	LuaVector2Object *pObj = (LuaVector2Object *)lua_touserdata(pLuaState, 1);
	if(!pObj)
	{
		return 0;
	}

	LuaVector2Object* pOther;
	if((pOther=reinterpret_cast<LuaVector2Object*>
			(luaL_checkudata(pLuaState,2,pObj->type()->name().c_str()))))
	{
		*pObj->data()=*pOther->data();
		t_lua_vector2::push(pObj->context(),pObj->data());
		return 1;
	}

	Real vector[2]={ 0.0f,0.0f };
	if(lua_isnumber(pLuaState, 2))
	{
		vector[0]=(Real)lua_tonumber(pLuaState, 2);
	}
	if(lua_isnumber(pLuaState, 3))
	{
		vector[1]=(Real)lua_tonumber(pLuaState, 3);
	}

	fe::set<Real>(*pObj->data(),vector[0],vector[1]);
	t_lua_vector2::push(pObj->context(),pObj->data());
	return 1;
}

void LuaVector2Type::pushNewMetaTable(void)
{
	lua_State *pLuaState = m_pContext->state();

//	lua_newtable(pLuaState);
	luaL_newmetatable(pLuaState,m_mtname.c_str());

	// call index newindex add sub mul div unm pow concat lt le eq

	lua_pushstring(pLuaState, "__gc");
	lua_pushcfunction(pLuaState, LuaVector2Object::destroy);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__index");
	lua_pushcfunction(pLuaState, LuaVector2Object::index);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__newindex");
	lua_pushcfunction(pLuaState, LuaVector2Object::newindex);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__eq");
	lua_pushcfunction(pLuaState, LuaVector2Object::eq);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__unm");
	lua_pushcfunction(pLuaState, LuaVector2Object::unm);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__add");
	lua_pushcfunction(pLuaState, LuaVector2Object::add);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__sub");
	lua_pushcfunction(pLuaState, LuaVector2Object::sub);
	lua_settable(pLuaState, -3);

	lua_pushstring(pLuaState, "__mul");
	lua_pushcfunction(pLuaState, LuaVector2Object::mul);
	lua_settable(pLuaState, -3);
}

} /* namespace */

