/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __lua_LuaObject_h__
#define __lua_LuaObject_h__

namespace fe
{

class LuaBaseObject
{
	public:
		LuaBaseObject(void) {}
virtual	~LuaBaseObject(void) {}
		sp<LuaBaseType >	&type(void) { return m_spType; }
		LuaContext			*context(void) { return m_spType->context(); }
	protected:
		sp<LuaBaseType >	m_spType;
};

template<class T, FE_UWORD tid>
class LuaObject : public LuaBaseObject
{
	private:
		LuaObject(void) { feX("LuaObject::LuaObject", "should never call"); }
	public:
		typedef LuaObject<T,tid> t_luott;
		LuaObject(LuaContext *pContext, const T &data)
		{
			m_spType = pContext->lookup(tid);
			m_data = data;
			m_spType->pushMetaTable();
			lua_setmetatable(m_spType->state(), -2);
		}
virtual	~LuaObject(void){}
		// factory that pushes onto stack
static	LuaObject<T,tid> *push(LuaContext *pContext, const T &data)
		{
			void *pBlock = lua_newuserdata(pContext->state(),
				sizeof(t_luott));
			t_luott *pObj = new(pBlock)t_luott(pContext, data);
			return pObj;
		}
		T	&data(void) { return m_data; }
static	int		destroy(lua_State *pLuaState)
		{
			void *pV = lua_touserdata(pLuaState, -1);
			t_luott *pW = (t_luott *)pV;
			pW->~t_luott();
			return 0;
		}
static	t_luott	*selfIsFirst(lua_State *pLuaState)
		{
			if(!lua_isuserdata(pLuaState, 1))
			{
				LuaContext::error(pLuaState, "self not first parameter");
				return NULL;
			}
			t_luott *pObj = (t_luott *)lua_touserdata(pLuaState, 1);
			if(pObj->type() != pObj->context()->lookup(tid))
			{
				LuaContext::error(pLuaState, "self not first parameter");
				return NULL;
			}
			return pObj;
		}

static	String	argType(lua_State *pLuaState,U32 index)
				{
					if (!lua_getmetatable(pLuaState, 2))
					{
						return "<no metatable>";
					}

					lua_rawget(pLuaState, LUA_REGISTRYINDEX);
					return lua_tostring(pLuaState, -1);
				}
	protected:
		String	aliasedName(String a_name)
				{
					LuaContext* pContext=context();
					return pContext? pContext->aliasedName(a_name): a_name;
				}

	private:
		T					m_data;
};

/* ----------------------------------------------------------------------- */
// RECORD
typedef LuaObject<Record, LuaContext::e_record> t_lua_record;
class LuaRecordObject : public t_lua_record
{
	public:
virtual				~LuaRecordObject(void) {}
static		int		index(lua_State *pLuaState);
static		int		newindex(lua_State *pLuaState);
static		int		isValid(lua_State *pLuaState);
static		int		check(lua_State *pLuaState);
static		int		layout(lua_State *pLuaState);
static		int		clone(lua_State *pLuaState);
//static		int		deepclone(lua_State *pLuaState);
};


typedef LuaType<Record, LuaRecordObject> t_lua_record_t;
class LuaRecordType : public t_lua_record_t
{
	public:
					LuaRecordType(LuaContext *pContext,const String &mtn);
virtual				~LuaRecordType(void){}
virtual		void	pushNewMetaTable(void);
};


/* ----------------------------------------------------------------------- */
// LAYOUT
typedef LuaObject<sp<Layout>, LuaContext::e_layout> t_lua_layout;
class LuaLayoutObject : public t_lua_layout
{
	public:
virtual				~LuaLayoutObject(void) {}
static		int		index(lua_State *pLuaState);
static		int		scope(lua_State *pLuaState);
static		int		populate(lua_State *pLuaState);
static		int		createRecord(lua_State *pLuaState);
static		int		name(lua_State *pLuaState);
};

typedef LuaType<sp<Layout>, LuaLayoutObject> t_lua_layout_t;
class LuaLayoutType : public t_lua_layout_t
{
	public:
					LuaLayoutType(LuaContext *pContext,const String &mtn);
virtual				~LuaLayoutType(void){}
};

/* ----------------------------------------------------------------------- */
// SCOPE
typedef LuaObject<sp<Scope>, LuaContext::e_scope> t_lua_scope;
class LuaScopeObject : public t_lua_scope
{
	public:
virtual				~LuaScopeObject(void) {}
static		int		index(lua_State *pLuaState);
static		int		declare(lua_State *pLuaState);
static		int		createRecord(lua_State *pLuaState);
};

typedef LuaType<sp<Scope>, LuaScopeObject> t_lua_scope_t;
class LuaScopeType : public t_lua_scope_t
{
	public:
					LuaScopeType(LuaContext *pContext,const String &mtn);
virtual				~LuaScopeType(void){}
};

/* ----------------------------------------------------------------------- */
// RECORD GROUP
typedef LuaObject<sp<RecordGroup>, LuaContext::e_group> t_lua_group;
class LuaGroupObject : public t_lua_group
{
	public:
virtual				~LuaGroupObject(void) {}
static		int		index(lua_State *pLuaState);
static		int		iterator(lua_State *pLuaState);
static		int		newiterator(lua_State *pLuaState);
static		int		length(lua_State *pLuaState);
static		int		add(lua_State *pLuaState);
static		int		remove(lua_State *pLuaState);
static		int		lookup(lua_State *pLuaState);
static		int		setWeak(lua_State *pLuaState);
static		int		clear(lua_State *pLuaState);
};

typedef LuaType<sp<RecordGroup>, LuaGroupObject> t_lua_group_t;
class LuaGroupType : public t_lua_group_t
{
	public:
					LuaGroupType(LuaContext *pContext,const String &mtn);
virtual				~LuaGroupType(void){}
virtual		void	pushNewMetaTable(void);
};

/* ----------------------------------------------------------------------- */
// RECORD GROUP ITERATOR
typedef LuaObject<RecordGroup::iterator, LuaContext::e_groupit> t_lua_groupit;
class LuaGroupItObject : public t_lua_groupit
{
	public:
virtual				~LuaGroupItObject(void) {}
static		int		index(lua_State *pLuaState);
};

typedef LuaType<RecordGroup::iterator, LuaGroupItObject> t_lua_groupit_t;
class LuaGroupItType : public t_lua_groupit_t
{
	public:
					LuaGroupItType(LuaContext *pContext,const String &mtn);
virtual				~LuaGroupItType(void){}
};

/* ----------------------------------------------------------------------- */
// RECORD Array
typedef LuaObject<sp<RecordArray>, LuaContext::e_array> t_lua_array;
class LuaArrayObject : public t_lua_array
{
	public:
virtual				~LuaArrayObject(void) {}
static		int		index(lua_State *pLuaState);
static		int		iterator(lua_State *pLuaState);
static		int		newiterator(lua_State *pLuaState);
static		int		length(lua_State *pLuaState);
static		int		check(lua_State *pLuaState);
static		int		add(lua_State *pLuaState);
static		int		remove(lua_State *pLuaState);
};

typedef LuaType<sp<RecordArray>, LuaArrayObject> t_lua_array_t;
class LuaArrayType : public t_lua_array_t
{
	public:
					LuaArrayType(LuaContext *pContext,const String &mtn);
virtual				~LuaArrayType(void){}
virtual		void	pushNewMetaTable(void);
};

class LuaArrayContext
{
	public:
			sp<RecordArray>	m_spRecordArray;
			I32				m_index;
};

typedef LuaObject<LuaArrayContext, LuaContext::e_arrayit> t_lua_arrayit;
class LuaArrayItObject : public t_lua_arrayit
{
	public:
virtual				~LuaArrayItObject(void) {}
static		int		index(lua_State *pLuaState);
};

typedef LuaType<LuaArrayContext, LuaArrayItObject> t_lua_arrayit_t;
class LuaArrayItType : public t_lua_arrayit_t
{
	public:
					LuaArrayItType(LuaContext *pContext,const String &mtn);
virtual				~LuaArrayItType(void){}
};

/* ----------------------------------------------------------------------- */
// integer array
typedef LuaObject< Array<I32>*, LuaContext::e_intarray> t_lua_intarray;
class LuaIntArrayObject : public t_lua_intarray
{
	public:
virtual				~LuaIntArrayObject(void)								{}
static		int		index(lua_State *pLuaState);
static		int		iterator(lua_State *pLuaState);
static		int		newiterator(lua_State *pLuaState);
static		int		length(lua_State *pLuaState);
static		int		add(lua_State *pLuaState);
static		int		resize(lua_State *pLuaState);
};

typedef LuaType< Array<I32>* , LuaIntArrayObject> t_lua_intarray_t;
class LuaIntArrayType : public t_lua_intarray_t
{
	public:
					LuaIntArrayType(LuaContext *pContext,const String &mtn);
virtual				~LuaIntArrayType(void)									{}
virtual		void	pushNewMetaTable(void);
};

class LuaIntArrayContext
{
	public:
			Array<I32>*	m_pIntArray;
			I32			m_index;
};

typedef LuaObject<LuaIntArrayContext,
	LuaContext::e_intarrayit> t_lua_intarrayit;
class LuaIntArrayItObject : public t_lua_intarrayit
{
	public:
virtual				~LuaIntArrayItObject(void)								{}
static		int		index(lua_State *pLuaState);
};

typedef LuaType<LuaIntArrayContext, LuaIntArrayItObject> t_lua_intarrayit_t;
class LuaIntArrayItType : public t_lua_intarrayit_t
{
	public:
					LuaIntArrayItType(LuaContext *pContext,const String &mtn);
virtual				~LuaIntArrayItType(void)								{}
};

/* ----------------------------------------------------------------------- */
// float vector
typedef LuaObject<float *, LuaContext::e_vector> t_lua_vector;
class LuaVectorObject : public t_lua_vector
{
	public:
virtual				~LuaVectorObject(void) {}
static		int		newindex(lua_State *pLuaState);
static		int		index(lua_State *pLuaState);
};

typedef LuaType<float *, LuaVectorObject> t_lua_vector_t;
class LuaVectorType : public t_lua_vector_t
{
	public:
					LuaVectorType(LuaContext *pContext,const String &mtn);
virtual				~LuaVectorType(void){}
virtual		void	pushNewMetaTable(void);
};

/* ----------------------------------------------------------------------- */
// Vector4 specialization
typedef LuaObject<Vector4*, LuaContext::e_vector4> t_lua_vector4;
class LuaVector4Object : public t_lua_vector4
{
	public:
virtual				~LuaVector4Object(void) {}
static		int		newindex(lua_State *pLuaState);
static		int		index(lua_State *pLuaState);
static		int		eq(lua_State *pLuaState);
static		int		unm(lua_State *pLuaState);
static		int		add(lua_State *pLuaState);
static		int		sub(lua_State *pLuaState);
static		int		mul(lua_State *pLuaState);
static		int		unit(lua_State *pLuaState);
static		int		normalize(lua_State *pLuaState);
static		int		magnitude(lua_State *pLuaState);
static		int		dot(lua_State *pLuaState);
static		int		cross3(lua_State *pLuaState);
static		int		set(lua_State *pLuaState);
};

typedef LuaType<Vector4*, LuaVector4Object> t_lua_vector4_t;
class LuaVector4Type : public t_lua_vector4_t
{
	public:
					LuaVector4Type(LuaContext *pContext,const String &mtn);
virtual				~LuaVector4Type(void){}
virtual		void	pushNewMetaTable(void);
};

/* ----------------------------------------------------------------------- */
// Vector4f specialization
typedef LuaObject<Vector4f*, LuaContext::e_vector4f> t_lua_vector4f;
class LuaVector4fObject : public t_lua_vector4f
{
	public:
virtual				~LuaVector4fObject(void) {}
static		int		newindex(lua_State *pLuaState);
static		int		index(lua_State *pLuaState);
static		int		eq(lua_State *pLuaState);
static		int		unm(lua_State *pLuaState);
static		int		add(lua_State *pLuaState);
static		int		sub(lua_State *pLuaState);
static		int		mul(lua_State *pLuaState);
static		int		unit(lua_State *pLuaState);
static		int		normalize(lua_State *pLuaState);
static		int		magnitude(lua_State *pLuaState);
static		int		dot(lua_State *pLuaState);
static		int		cross3(lua_State *pLuaState);
static		int		set(lua_State *pLuaState);
};

typedef LuaType<Vector4f*, LuaVector4fObject> t_lua_vector4f_t;
class LuaVector4fType : public t_lua_vector4_t
{
	public:
					LuaVector4fType(LuaContext *pContext,const String &mtn);
virtual				~LuaVector4fType(void){}
virtual		void	pushNewMetaTable(void);
};

/* ----------------------------------------------------------------------- */
// Vector3 specialization
typedef LuaObject<Vector3*, LuaContext::e_vector3> t_lua_vector3;
class LuaVector3Object : public t_lua_vector3
{
	public:
virtual				~LuaVector3Object(void) {}
static		int		newindex(lua_State *pLuaState);
static		int		index(lua_State *pLuaState);
static		int		eq(lua_State *pLuaState);
static		int		unm(lua_State *pLuaState);
static		int		add(lua_State *pLuaState);
static		int		sub(lua_State *pLuaState);
static		int		mul(lua_State *pLuaState);
static		int		unit(lua_State *pLuaState);
static		int		normalize(lua_State *pLuaState);
static		int		magnitude(lua_State *pLuaState);
static		int		dot(lua_State *pLuaState);
static		int		cross3(lua_State *pLuaState);
static		int		set(lua_State *pLuaState);
static		int		tostring(lua_State *pLuaState);
};

typedef LuaType<Vector3*, LuaVector3Object> t_lua_vector3_t;
class LuaVector3Type : public t_lua_vector3_t
{
	public:
					LuaVector3Type(LuaContext *pContext,const String &mtn);
virtual				~LuaVector3Type(void){}
virtual		void	pushNewMetaTable(void);
};

/* ----------------------------------------------------------------------- */
// Vector2 specialization
typedef LuaObject<Vector2*, LuaContext::e_vector3> t_lua_vector2;
class LuaVector2Object : public t_lua_vector2
{
	public:
virtual				~LuaVector2Object(void) {}
static		int		newindex(lua_State *pLuaState);
static		int		index(lua_State *pLuaState);
static		int		eq(lua_State *pLuaState);
static		int		unm(lua_State *pLuaState);
static		int		add(lua_State *pLuaState);
static		int		sub(lua_State *pLuaState);
static		int		mul(lua_State *pLuaState);
static		int		unit(lua_State *pLuaState);
static		int		normalize(lua_State *pLuaState);
static		int		magnitude(lua_State *pLuaState);
static		int		dot(lua_State *pLuaState);
static		int		set(lua_State *pLuaState);
};

typedef LuaType<Vector2*, LuaVector2Object> t_lua_vector2_t;
class LuaVector2Type : public t_lua_vector2_t
{
	public:
					LuaVector2Type(LuaContext *pContext,const String &mtn);
virtual				~LuaVector2Type(void){}
virtual		void	pushNewMetaTable(void);
};


} /* namespace */

#endif /* __lua_LuaObject_h__ */

