/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __usd_DrawHydra_h__
#define __usd_DrawHydra_h__

namespace fe
{

/**************************************************************************//**
	@brief Hydra-specific replacements for DrawI

	@ingroup usd

	Features not implemented using Hydra may fall back on DrawOpenGL.
*//***************************************************************************/
class FE_DL_EXPORT DrawHydra: public DrawOpenGL,
		public Initialize<DrawHydra>
{
	public:
						DrawHydra();
virtual					~DrawHydra();

		void			initialize(void);

virtual	void			flush(void);
virtual	sp<DrawBufferI>	createBuffer(void);

virtual	void			drawPoints(const SpatialVector *vertex,
								const SpatialVector *normal,U32 vertices,
								BWORD multicolor,const Color *color,
								sp<DrawBufferI> spDrawBuffer);

virtual	void			drawTriangles(const SpatialVector *vertex,
								const SpatialVector *normal,
								const Vector2f *texture,U32 vertices,
								StripMode strip,BWORD multicolor,
								const Color *color,
								const Array<I32>* hullPointMap,
								const Array<Vector4i>* hullFacePoint,
								sp<DrawBufferI> spDrawBuffer);

	protected:

virtual	void			drawLinesInternal(const SpatialVector *vertex,
								const SpatialVector *normal,
								const Vector2f *texture,U32 vertices,
								StripMode strip,BWORD multicolor,
								const Color *color,
								const Vector3i *element,U32 elementCount,
								sp<DrawBufferI> spDrawBuffer);

virtual	void			drawPolys(U32 grain,const SpatialVector *vertex,
								const SpatialVector *normal,
								const Vector2f *texture,
								U32 vertices,StripMode strip,
								BWORD multicolor,const Color *color,
								const Array<I32>* hullPointMap,
								const Array<Vector4i>* hullFacePoint,
								sp<DrawBufferI> spDrawBuffer);

	class Buffer: public DrawOpenGL::Buffer
	{
		public:
						Buffer(void);
	virtual				~Buffer(void);
			void		releaseAll(void);

			HydraDriver*	m_pDriver;
			GLuint			m_vao;
			BWORD			m_multicolor;
			Color			m_firstColor;
	};

	private:

		void			updateDriver(HydraDriver*& a_rpDriver);
		void			updateLighting(HydraDriver* a_pDriver);

#if FE_HYDRA_TEST_LIGHTING
		TestLightingShaderSharedPtr			m_lightingShader;
#endif

		Real			m_alphaMin;
		I32				m_shadowResolution;
		Real			m_shadowBias;
};

} // namespace

#endif /* __usd_DrawHydra_h__ */
