//
// Copyright 2016 Pixar
//
// Licensed under the Apache License, Version 2.0 (the "Apache License")
// with the following modification; you may not use this file except in
// compliance with the Apache License and the following modification to it:
// Section 6. Trademarks. is deleted and replaced with:
//
// 6. Trademarks. This License does not grant permission to use the trade
//	names, trademarks, service marks, or product names of the Licensor
//	and its affiliates, except as required to comply with Section 4(c) of
//	the License and to reproduce the content of the NOTICE file.
//
// You may obtain a copy of the Apache License at
//
//	 http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the Apache License with the above modification is
// distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied. See the Apache License for the specific
// language governing permissions and limitations under the Apache License.

// originally from pxr/imaging/lib/hd/unitTestHelper.h
// adapted for Free Electron

#ifndef TEST_LIGHTING_SHADER
#define TEST_LIGHTING_SHADER

#include "pxr/imaging/hd/engine.h"
#include "pxr/imaging/hd/lightingShader.h"
#include "pxr/imaging/hd/renderPass.h"
#include "pxr/imaging/hd/renderPassState.h"
#include "pxr/imaging/glf/glslfx.h"

#include "pxr/base/gf/vec4d.h"
#include "pxr/base/gf/matrix4d.h"

#include <vector>
#include <boost/scoped_ptr.hpp>

/// \class TestLightingShader
///
/// A custom lighting shader for unit tests.
///
typedef boost::shared_ptr<class TestLightingShader> TestLightingShaderSharedPtr;

class TestLightingShader : public HdLightingShader {
public:
	TestLightingShader();
	virtual ~TestLightingShader();

	/// HdShader overrides
	virtual ID ComputeHash() const;
	virtual std::string GetSource(TfToken const &shaderStageKey) const;
	virtual void BindResources(Hd_ResourceBinder const &binder, int program);
	virtual void UnbindResources(Hd_ResourceBinder const &binder, int program);
	virtual void AddBindings(HdBindingRequestVector *customBindings);

	/// HdLightingShader overrides
	virtual void SetCamera(GfMatrix4d const &worldToViewMatrix,
						   GfMatrix4d const &projectionMatrix);

	void SetSceneAmbient(GfVec3f const &color);
	void SetLight(int light, GfVec3f const &dir, GfVec3f const &color);

private:
	struct Light {
		GfVec3f dir;
		GfVec3f eyeDir;
		GfVec3f color;
	};
	Light _lights[2];
	GfVec3f _sceneAmbient;
	boost::scoped_ptr<GlfGLSLFX> _glslfx;
};

#endif  // TEST_LIGHTING_SHADER
